from wsgiref.util import request_uri

from elasticapm.conf import constants
from elasticapm.utils import get_url_dict
from elasticapm.utils.wsgi import get_environ, get_headers


def get_data_from_request(request, config, event_type):
    result = {
        "env": dict(get_environ(request.environ)),
        "method": request.method,
        "socket": {"remote_address": request.environ.get("REMOTE_ADDR")},
    }
    if config.capture_headers:
        result["headers"] = dict(get_headers(request.environ))
    if request.method.upper() in constants.HTTP_WITH_BODY:
        if config.capture_body not in ("all", event_type):
            result["body"] = "[REDACTED]"
        else:
            result["body"] = request.content

    result["url"] = get_url_dict(request_uri(request.environ))

    return result


def get_data_from_response(response, config, event_type):
    return {
    	'status_code': response.status_code,
    	'headers': dict(response.headers)
    }



from .specification import Specification
from .router import Router
from .authorizer import Authorizer
from .authorizer import ScopesFetcher
from .validator import Validator
from .operator import Operator
from .operator import default_operation_id
from .jsonschema import schematize


from .. import jwt
from ..exception import Unauthorized
from ..exception import Forbidden


class ScopesFetcher:

    def fetch_scopes(self, request, security, required_scopes):
        if security == 'fermata.jwt':
            headers = [('x-required-scopes', ','.join(required_scopes))]
            try:
                authorized = bool(request.identity)
                scopes = request.identity.scopes
                return authorized, scopes
            except jwt.ExpiredToken:
                raise Unauthorized('the token is expired', headers=headers)
            except jwt.InvalidToken:
                raise Unauthorized('invalid token', headers=headers)
        return True, required_scopes


class Authorizer:

    def __init__(self, securities, scopes_fetcher=None):
        self.securities = securities
        self.scopes_fetcher = scopes_fetcher

    def authorize(self, request):
        for s in  self.securities:
            for security, required_scopes in s.items():
                required_scopes = set(required_scopes)
                authorized, scopes = self.scopes_fetcher.fetch_scopes(
                    request, security, required_scopes)
                scopes = set(scopes)
                headers = [
                    ('x-required-scopes', ','.join(required_scopes)), 
                    ('x-gained-scopes', ','.join(scopes))]
                if not authorized:
                    raise Unauthorized(headers=headers)
                if not required_scopes.issubset(scopes):
                    raise Forbidden(
                        f'require scopes: ({" ".join(required_scopes)}), '
                        f'gained scopes: ({" ".join(scopes)})',
                        headers=headers)
                return headers
        return []


import typing
import sys
import glob
import os.path
import traceback

from . import loader
from .jwt import JWT
from .request import Request
from .response import Response
from .response import JSONResponse
from .openapi import Router
from .openapi import Authorizer
from .openapi import Validator
from .openapi import Operator
from .openapi import Specification
from .openapi import default_operation_id
from .openapi import ScopesFetcher
from .openapi import schematize
from .exception import HTTPException
from .exception import InternalServerError

MAX_CONTENT_LENGTH = 1024 * 10
ALLOWED_OPERATIONS = {'get', 'put', 'post', 'delete', 'options', 'head', 'patch', 'trace'}
PAYLOAD_OPERATIONS = {'post', 'put', 'options'}


def respond_abort(request, exception: HTTPException):
    return request.app.response_class(**exception)


def respond_capture(request, exception):
    traceback.print_exc()
    return respond_abort(request, InternalServerError())


class Fermata(ScopesFetcher):

    def __init__(
            self,
            default_package: str = None,
            spec_glob: str = None,
            *,
            exception_captors: typing.Dict[typing.Type[Exception], typing.Callable] = None,
            scopes_fetcher: typing.Callable = None,
            jwt_key: str = None,
            default_schematize = None,
            response_class = None,
        ) -> None:

        self.default_package = default_package
        if scopes_fetcher:
            self.fetch_scopes = scopes_fetcher
        self.listeners = {}
        self.default_schematize = default_schematize
        self.response_class = response_class or JSONResponse
        if jwt_key:
            self.jwt = JWT(jwt_key)
        self.max_content_length = MAX_CONTENT_LENGTH

        self.router = Router()
        self.spec_files = []
        if spec_glob:
            self.scan(spec_glob)

        self.exception_captors = {
            HTTPException: respond_abort,
            Exception: respond_capture,
            **(exception_captors or {})
        }

    @property
    def name(self):
        return self.default_package

    def preload(self):
        for _, methods in self.router:
            for method, data in methods.items():
                if 'operator' not in data:
                    continue
                data.get('operator').load()

    def scan(self, spec_glob, filename_as_basepath=True):
        spec_files = {}
        parsers = {
            '.json': 'json.load',
            '.yaml': 'yaml.safe_load',
            '.yml': 'yaml.safe_load',
        }
        for path in glob.glob(spec_glob):
            ext = os.path.splitext(path)[1]
            if ext not in parsers:
                self.log('warning', f'not supported file: {path}')
                continue
            spec_files.setdefault(parsers[ext], []).append(path)
        for parser, files in spec_files.items():
            parse = loader.load(parser)
            for f in files:
                spec = parse(open(f))
                base_path = ''
                if filename_as_basepath:
                    base_path = os.path.splitext(os.path.basename(f))[0]
                self.interpret(spec, base_path)
                self.spec_files.append(f)
        return self

    def interpret(self, spec, base_path=''):
        spec = Specification(spec)

        # router
        for expression in spec.data.get('paths', {}).keys():
            self.router.add(base_path + expression)

        # authorizer
        # securitySchemes
        security = spec.ref('components/securitySchemes')
        if security:
            base = spec.ref('security')
            for k, node in spec.items('paths/{path}/{operation}'):
                security = node.get('security', base)
                if not security:
                    continue
                path, operation = k.values()
                authorizer = Authorizer(security, self)
                self.router.set(authorizer, base_path + path, operation)

        # validator
        # parameters as validate clauses
        for k, node in spec.items('paths/{path}/{operation}'):
            path, operation = k.values()
            if operation not in ALLOWED_OPERATIONS:
                continue
            p = path.replace('/', '~1')
            base = spec.ref(f'paths/{p}/parameters', [])
            parameters = base + node.get('parameters', [])
            if not parameters and 'requestBody' not in node:
                continue
            body = node.get('requestBody', {})
            validator = Validator(
                parameters, 
                body.get('content', {}),
                body.get('required', False))
            self.router.set(validator, base_path + path, operation)

        # operator
        for k, node in spec.items('paths/{path}/{operation}'):
            path, operation = k.values()
            if operation not in ALLOWED_OPERATIONS:
                continue
            full_path = base_path + path
            op = node.get('operationId') or default_operation_id(full_path, operation)
            response_schema = spec.ref(
                f'paths/{path.replace("/", "~1")}/{operation}/'
                f'responses/200/content/'
                f'{self.response_class.content_type.replace("/", "~1")}/schema')
            operator = Operator(op, self.default_package, response_schema)
            self.router.set(operator, full_path, operation)

        return self

    def respond(self, request):
        authorizer, validator, operator = self.router.route(request)
        headers = authorizer.authorize(request) if authorizer else []
        params = validator.validate(request) if validator else {}
        res = operator.operate(request, **params)
        if not isinstance(res, Response):
            if isinstance(res, tuple):
                try:
                    if isinstance(res[1], list):
                        res = res[0], 200, res[1]
                except IndexError:
                    raise TypeError(
                        "operator return tuple form must be one of: "
                        "(body, status, headers), (body, status) or "
                        "(body, headers).")
                res = list(res)
            else:
                res = [res]
            if operator.schema:
                res[0] = schematize(operator.schema, res[0], self.default_schematize)
            res = self.response_class(*res[:3])
        res.headers += headers
        return res

    def abort(self, request, exception):
        return self.capture(request, exception)

    def capture(self, request, exception):
        for cls in type(exception).__mro__:
            if cls in self.exception_captors:
                return self.exception_captors[cls](request, exception)
        return respond_capture(request, exception)

    def identity(self, request):
        parts = request.headers.get('authorization', '').split(maxsplit=1)
        if not parts or not self.jwt:
            return None
        token = self.jwt.loads(parts.pop())
        return token

    def on(self, event, listener):
        self.listeners.setdefault(event, []).append(listener)

    def off(self, event, listener):
        try:
            self.listeners.get(event, []).remove(listener)
        except ValueError:
            pass

    def trigger(self, event, *args, **kwargs):
        if event in self.listeners:
            for l in self.listeners[event]:
                l(*args, **kwargs)

    def log(self, *args, **kwargs):
        self.trigger('log', *args, **kwargs)

    def wsgi(
            self,
            environ: dict,
            start_response: typing.Callable
        ) -> typing.Sequence:
        try:
            environ['fermata.application'] = self
            request = Request(environ)
            self.trigger('request', request=request)
            response = self.respond(request)
            self.trigger('respond', request=request, response=response)
        except HTTPException as e:
            response = self.abort(request, e)
            self.trigger('abort', request=request, response=response)
        except Exception as e:
            response = self.capture(request, e)
            self.trigger('capture', request=request, response=response)

        return response(environ, start_response)

    def __call__(
            self,
            environ: dict,
            start_response: typing.Callable
        ) -> typing.Sequence:
        return self.wsgi(environ, start_response)


try:
    from functools import cached_property
except ImportError:


    class cached_property:
        '''A thread unsafe cached_property implementation.
        '''

        def __init__(self, func):
            self.func = func
            self.__doc__ = func.__doc__

        def __set_name__(self, _, name):
            self.__name__ = name

        def __get__(self, instance, _=None):
            if instance is None:
                return self
            instance.__dict__[self.__name__] = self.func(instance)

            return instance.__dict__[self.__name__]

try:
    from functools import singledispatchmethod
except ImportError:


    class singledispatchmethod:
        '''A poor implementation for type dispatch.
           WARNING: this is not an equivalent `singledispatchmethod` in python 3.8
        '''

        def __init__(self, func):
            self.func = func
            self.dispatcher = {}

        def register(self, cls):
            def wrap(method):
                self.dispatcher[cls] = method
                return method

            return wrap

        def __get__(self, obj, cls=None):
            def _method(*args, **kwargs):
                method = self.dispatcher.get(args[0].__class__, self.func)
                return method.__get__(obj, cls)(*args, **kwargs)

            _method.__name__ = self.func.__name__
            _method.__doc__ = self.func.__doc__

            return _method



import typing
import types
import orjson

HTTP_STATUS_PHRASE = {
    100: "CONTINUE",
    101: "SWITCHING PROTOCOLS",
    102: "PROCESSING",
    200: "OK",
    201: "CREATED",
    202: "ACCEPTED",
    203: "NON AUTHORITATIVE INFORMATION",
    204: "NO CONTENT",
    205: "RESET CONTENT",
    206: "PARTIAL CONTENT",
    207: "MULTI STATUS",
    208: "ALREADY REPORTED",
    226: "IM USED",
    300: "MULTIPLE CHOICES",
    301: "MOVED PERMANENTLY",
    302: "FOUND",
    303: "SEE OTHER",
    304: "NOT MODIFIED",
    305: "USE PROXY",
    307: "TEMPORARY REDIRECT",
    308: "PERMANENT REDIRECT",
    400: "BAD REQUEST",
    401: "UNAUTHORIZED",
    402: "PAYMENT REQUIRED",
    403: "FORBIDDEN",
    404: "NOT FOUND",
    405: "METHOD NOT ALLOWED",
    406: "NOT ACCEPTABLE",
    407: "PROXY AUTHENTICATION REQUIRED",
    408: "REQUEST TIMEOUT",
    409: "CONFLICT",
    410: "GONE",
    411: "LENGTH REQUIRED",
    412: "PRECONDITION FAILED",
    413: "REQUEST ENTITY TOO LARGE",
    414: "REQUEST URI TOO LONG",
    415: "UNSUPPORTED MEDIA TYPE",
    416: "REQUESTED RANGE NOT SATISFIABLE",
    417: "EXPECTATION FAILED",
    421: "MISDIRECTED REQUEST",
    422: "UNPROCESSABLE ENTITY",
    423: "LOCKED",
    424: "FAILED DEPENDENCY",
    426: "UPGRADE REQUIRED",
    428: "PRECONDITION REQUIRED",
    429: "TOO MANY REQUESTS",
    431: "REQUEST HEADER FIELDS TOO LARGE",
    451: "UNAVAILABLE FOR LEGAL REASONS",
    500: "INTERNAL SERVER ERROR",
    501: "NOT IMPLEMENTED",
    502: "BAD GATEWAY",
    503: "SERVICE UNAVAILABLE",
    504: "GATEWAY TIMEOUT",
    505: "HTTP VERSION NOT SUPPORTED",
    506: "VARIANT ALSO NEGOTIATES",
    507: "INSUFFICIENT STORAGE",
    508: "LOOP DETECTED",
    510: "NOT EXTENDED",
    511: "NETWORK AUTHENTICATION REQUIRED",
}


def status_code_phrase(status_code):
    phrase = HTTP_STATUS_PHRASE.get(status_code, 'UNKNOWN')
    return '%d %s' % (status_code, phrase)


class Response(object):

    charset = 'utf-8'
    content_type = 'text/plain'

    def __init__(
            self,
            body: typing.Any,
            status_code: int = 200,
            headers: typing.Sequence[typing.Tuple[str, str]] = None,
            content_type=None
        ):
        self.body = body
        self.status_code = status_code
        self.headers = headers or []
        if content_type:
            self.content_type = content_type

    def bytes(self) -> typing.Sequence:
        return self.body.encode(self.charset)

    def __call__(
            self,
            environ: dict,
            start_response: typing.Callable
        ) -> typing.Sequence:
        body = self.bytes()
        status = status_code_phrase(self.status_code)
        headers = self.headers + [(
                'content-type', 
                f'{self.content_type};charset={self.charset}')]

        start_response(status, [(str(k), str(v)) for k, v in headers])
        return [body, b'\n']


class JSONResponse(Response):

    content_type = 'application/json'
    ORJSON_OPTION = 0

    def bytes(self) -> typing.Sequence:
        return orjson.dumps(
                self.body,
                option=self.ORJSON_OPTION
            )



import typing


class HTTPException(Exception):

    status_code = None
    error_code = None
    message = 'HTTP Exception'

    _headers = []

    def __init__(self,
            message: str = None,
            error_code: str = None,
            headers: typing.Sequence[typing.Tuple[str, str]] = None
        ) -> None:
        if message is not None:
            self.message = message
        if error_code is not None:
            self.error_code = error_code
        if headers is not None:
            self._headers = headers

    @property
    def body(self):
        return { k: getattr(self, k) 
            for k in self.__class__.__dict__
            if not k.startswith('_') }

    @property
    def headers(self):
        return self._headers

    def keys(self):
        return ['status_code', 'body', 'headers']

    def __getitem__(self, k):
        return getattr(self, k)


class BadRequest(HTTPException):

    status_code = 400
    error_code = 'bad_request'
    message = 'Bad Request'


class JSONDecodeError(BadRequest):

    error_code = 'json_decode_error'
    message = 'JSON DecodeError'


class Unauthorized(BadRequest):

    status_code = 401
    error_code = 'unauthorized'
    message = 'Unauthorized'


class Forbidden(BadRequest):

    status_code = 403
    error_code = 'forbidden'
    message = 'Forbidden'


class NotFound(BadRequest):

    status_code = 404
    error_code = 'path_not_found'
    message = 'Request Path Not Found'


class UnprocessableEntity(BadRequest):

    status_code = 422
    error_code = 'unprocessable_entity'
    message = 'Unprocessable Entity'


class VerificationFailed(UnprocessableEntity):

    error_code = 'verification_failed'
    message = 'Verification Failed'
    details = None

    def __init__(
            self,
            details: typing.Sequence[typing.Dict[str, str]],
            message: str = None,
            error_code: str = None,
            headers: typing.Sequence[typing.Tuple[str, str]] = None
        ) -> None:
        super(VerificationFailed, self).__init__(message, error_code, headers)
        self.details = details


class PathNotFound(NotFound):

    error_code = 'path_not_found'
    message = 'Request Path Not Found'


class PathNotFound(NotFound):

    error_code = 'path_not_found'
    message = 'Request Path Not Found'


class MethodNotAllowed(BadRequest):

    status_code = 405
    error_code = 'method_not_allowed'
    message = 'Method Not Allowed'        


class InternalServerError(HTTPException):

    status_code = 500
    error_code = 'internal_server_error'
    message = 'Internal Server Error'        


class BadGateway(InternalServerError):

    status_code = 502
    error_code = 'bad_gateway'
    message = 'Bad Gateway'        


class BadOperationId(BadGateway):

    error_code = 'bad_operation_id'
    message = 'Bad Operation Id'        



from fermata import __version__
import elasticapm
from elasticapm.conf import constants

from .util import get_data_from_request
from .util import get_data_from_response

'''
ref: [Creating Custom Framework Integrations with the Elastic APM Python Agent](https://www.elastic.co/cn/blog/creating-custom-framework-integrations-with-the-elastic-apm-python-agent)
'''


class ElasticAPM:

    def __init__(self, app, **defaults):
        elasticapm.instrument()
        self.client = elasticapm.Client(
            framework_name='fermata',
            framework_version=__version__,
            **defaults)
        app.on('request', self.request)
        app.on('respond', self.respond)
        app.on('abort', self.abort)
        app.on('capture', self.capture)
        app.apm = self

    def request(self, request):
        parent = elasticapm.trace_parent_from_headers(request.headers)
        self.client.begin_transaction('request', trace_parent=parent)
        elasticapm.set_context(
            lambda: get_data_from_request(
                request, self.client.config, constants.TRANSACTION),
            "request")

    def respond(self, request, response):
        elasticapm.set_context(
            lambda: get_data_from_response(
                response, self.client.config, constants.TRANSACTION),
            "request")
        name = f'{request.operation} {request.expression}'
        result = str(response.status_code)[:1] + 'xx'
        self.client.end_transaction(name, result)

    def abort(self, request, response):
        self.respond(request, response)

    def capture(self, request, response):
        self.client.capture_exception(
            context={
                "request": get_data_from_request(
                    request, self.client.config, constants.ERROR)
            },
            handled=False)
        self.respond(request, response)

    def log(self, *args, **kwargs):
        return self.client.capture_message(*args, **kwargs)


def init(app, **defaults):
    return ElasticAPM(app, **defaults)



from urllib.parse import parse_qs

import orjson

from . import jsonschema
from ..exception import VerificationFailed

CONTENT_NAME = 'body'


class FormDeserializer:

    def __init__(self, schema, encoding):
        self.encoding = encoding
        self.properties = {}
        for p, s in schema.get('properties', {}).items():
            type_ = s.get('items', s).get('type')
            self.properties[p] = {
                'multi': 'items' in s,
                'deserializer': DESERIALIZERS.get(type_)
            }

    def __call__(self, content: bytes):
        result = {}
        data = parse_qs(content)
        for p, clause in self.properties.items():
            multi = clause['multi']
            deserializer = clause['deserializer']
            try:
                result[p] = (next, list)[int(multi)](
                    deserializer(d) for d in data[p])
            except KeyError:
                continue
        return result

DESERIALIZERS = {
    'integer': int,
    'string': str,
    'number': float,
    'bool': lambda x: x in ('true', 'True'),
}

DESERIALIZERS_LIST = {
    'integer': lambda l: [int(v) for v in l],
    'string': lambda l: [str(v) for v in l],
    'number': lambda l: [float(v) for v in l],
    'bool': lambda l: [v in ('true', 'True') for v in l],
}

DESERIALIZERS_CONTENT = {
    'application/json': lambda schema, encoding: orjson.loads,
    'application/x-www-form-urlencoded': FormDeserializer,
}


def build_clause(param):
    location, name, required, schema = (
        param.get('in'),
        param.get('name'),
        param.get('required', False),
        param.get('schema', {}),
    )
    if not location or not name:
        return None

    # match request.attr
    location = {
        'query': 'params',
        'path': 'sections',
        'header': 'headers',
    }.get(location, location)

    # required no defaults
    default = None if required else schema.get('default', None)

    # > items MUST be present if the type is array.
    multi = 'items' in schema
    type_ = schema.get('items', schema).get('type')

    # type convert
    deserializer = (DESERIALIZERS, DESERIALIZERS_LIST)[int(multi)].get(type_)
    if 'content' in param:
        for content_type, node in param['content'].items():
            try:
                cls = DESERIALIZERS_CONTENT[content_type]
            except KeyError:
                # TODO: warning on console
                break
            schema = node.get('schema')
            encoding = node.get('encoding')
            # <Media Type Object>.encoding
            # need_encode = (content_type.startswith('multipart/') or 
            #     content_type == 'application/x-www-form-urlencoded')
            deserializer = cls(schema, encoding)
            break  # only one content type will be use

    # validator
    validator = jsonschema.create(schema) if schema else None

    return {
        'location': location,
        'name': name,
        'multi': multi,
        'required': required,
        'default': default,
        'type_': type_,
        'deserializer': deserializer,
        'validator': validator,
    }


def content_clause(content_type, param, required):
    '''
       :content_type: str
       :param: operation.requestBody.content[content_type] <Media Type Object>
    '''
    return build_clause({
        'required': required,
        'in': 'content',
        'name': CONTENT_NAME,
        'content': {
            content_type: {
                'schema': param.get('schema'),
                'encoding': param.get('encoding'),
            }
        }
    })


class Validator:

    def __init__(self, parameters, content, content_required=False):
        clauses = [build_clause(p) for p in parameters]
        self.clauses = [c for c in clauses if c]
        self.content_clauses = {}
        self.content_required = content_required
        for media, data in content.items():
            c = content_clause(media, data, content_required)
            if c:
                self.content_clauses[media] = c

    def _clauses(self, request):
        t = request.content_type or 'application/json'
        c = self.content_clauses.get(t) or self.content_clauses.get('*/*')
        if self.content_required and not c:
            c = { 'error': dict(name='content type', value=t, message='not acceptable') }
        return self.clauses + ([c] if c else [])

    def validate(self, request):
        params, errors = {}, []

        for clause in self._clauses(request):
            if 'error' in clause:
                errors.append(clause['error'])
                continue
            location = clause['location']
            name = clause['name']
            multi = clause['multi']
            default = clause['default']
            required = clause['required']

            data = getattr(request, location)
            if location == 'content':
                exist = len(data) > 0
            else:
                exist = name in data

            if location == 'content':
                value = data
            elif multi:
                value = data.get_list(name)
            else:
                value = data.get(name)

            # deserialize
            deserializer = clause['deserializer']
            if exist and deserializer:
                try:
                    value = deserializer(value)
                except ValueError:
                    errors.append(dict(name=name, value=value, message='type convert error'))

            # required and default
            if required and not exist:
                errors.append(dict(
                        name=name, 
                        message=f'var {name} in {location} is required'))
            elif default and not exist:
                exist, value = True, default

            # validate
            validator = clause['validator']
            if exist:
                if validator:
                    for e in validator.iter_errors(value):
                        errors.append(dict(name=name, value=value, message=e.message))
            elif required:
                errors.append(dict(name=name, message=f'{name} in {location} is required'))

            # store
            if exist:
                params[name] = value

        if errors:
            raise VerificationFailed(errors)

        return params



from .application import Fermata
from .request import Request
from .response import Response
from .response import JSONResponse
from .exception import HTTPException
from .exception import BadRequest
from .exception import JSONDecodeError
from .exception import Unauthorized
from .exception import Forbidden
from .exception import NotFound
from .exception import UnprocessableEntity
from .exception import VerificationFailed
from .exception import PathNotFound
from .exception import MethodNotAllowed
from .exception import InternalServerError
from .exception import BadGateway
from .exception import BadOperationId



from urllib.parse import parse_qs

from .future import cached_property
from .exception import JSONDecodeError
from .exception import Unauthorized
from . import jwt

HEADERS_WITHOUT_HTTP = ('CONTENT_TYPE', 'CONTENT_LENGTH')


def fmt_header_http(name):
    # header name is case-insensitive, 
    # people prefer Pascal Case in tradition,
    # but lowercase will be fashionable now.
    # https://tools.ietf.org/html/rfc7540#section-8.1.2
    return name.lower().replace('_', '-')


def fmt_header_wsgi(name):
    name = name.upper().replace("-", "_")
    prefix = ['HTTP_', ''][int(name in HEADERS_WITHOUT_HTTP)]
    return prefix + name


class Headers(object):

    def __init__(self, environ):
        self.environ = environ

    def get(self, name, default=None):
        k = fmt_header_wsgi(name)
        return self.environ.get(k, default)

    def __iter__(self):
        for key in self.environ:
            if key[:5] == 'HTTP_':
                yield fmt_header_http(key[5:])
            if key in HEADERS_WITHOUT_HTTP:
                yield fmt_header_http(key)

    def __contains__(self, key):
        return fmt_header_wsgi(key) in self.environ


class MultipleValues(object):

    def __init__(self, kv):
        self.kv = kv

    def get(self, name, default=None):
        return self.kv.get(name, [default])[0]

    def get_list(self, name):
        return self.kv.get(name, [])

    def __iter__(self):
        yield from self.kv

    def __contains__(self, key):
        return key in self.kv


class Identity:

    def __init__(self, token=None):
        if not token:
            token = {}
            self.authorized = False
        else:
            self.authorized = True
        self.id = token.get('sub')
        self.scopes = token.get('scopes', [])
        self.data = token.get('data', {})

    def __bool__(self):
        return self.authorized


class Request(object):

    def __init__(self, environ):
        self.environ = environ

    @property
    def app(self):
        return self.environ.get('fermata.application')

    @property
    def expression(self):
        return self.environ.get('fermata.expression')

    @property
    def sections(self):
        return self.environ.get('fermata.sections', {})

    @property
    def operation(self):
        return self.environ.get('fermata.operation', self.method)

    @property
    def content_type(self):
        t = self.environ.get('CONTENT_TYPE', '').lower().split(';')[0]
        return t or 'application/json'

    @property
    def path(self):
        return self.environ.get('PATH_INFO', '')

    @property
    def method(self):
        return self.environ.get('REQUEST_METHOD', 'get').lower()

    @property
    def remote_addr(self):
        return self.environ.get("REMOTE_ADDR", '')

    @cached_property
    def identity(self):
        return Identity(self.app.identity(self))

    @cached_property
    def headers(self):
        return Headers(self.environ)

    @cached_property
    def content(self):
        try:
            length = int(self.environ.get('CONTENT_LENGTH', -1))
        except ValueError:
            length = 0
        if self.app and (length > self.app.max_content_length or length < 0):
            length = self.app.max_content_length

        return self.environ['wsgi.input'].read(length).decode(errors='replace')

    @cached_property
    def params(self):
        return MultipleValues(parse_qs(
            self.environ.get('QUERY_STRING', ''), keep_blank_values=True))

    @cached_property
    def cookies(self):
        return {}  # not supported



import datetime
import uuid

from jsonschema import Draft7Validator
from jsonschema import validators
try:
    from numpy import ndarray
except ImportError:
    ndarray = list

SUPPORTED_TYPES = {
    None: (str, dict, list, tuple, int, float, bool,
        datetime.datetime, datetime.date, datetime.time, uuid.UUID, ndarray, None),
    'boolean': (bool, ),
    'object': (dict, ),
    'array': (list, set, tuple),
    'number': (float, ),
    'string': (str, datetime.datetime, datetime.date, datetime.time, uuid.UUID),
    'integer': (int),
    }


def extend_with_default(validator_class):
    '''copy from jsonschema docs'''

    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for property, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(property, subschema["default"])

        for error in validate_properties(
            validator, properties, instance, schema,
        ):
            yield error

    return validators.extend(
        validator_class, {"properties" : set_defaults},
    )

JSONSchema = extend_with_default(Draft7Validator)


class SchematizeError(ValueError):
    pass


class NullRefResolver:

    def push_scope(self, _):
        pass

    def pop_scope(self):
        pass

    def resolve(self, _):
        raise NotImplementedError('should not be called')

null_ref_resolver = NullRefResolver()


def create(schema):
    return JSONSchema(schema, resolver=null_ref_resolver)


def schematizable(obj, type_name):
    supported_types = SUPPORTED_TYPES.get(type_name)
    return supported_types and isinstance(obj, supported_types)


def schematize(schema, obj, default=None, keys=None):
    if keys is None:
        keys = []
    type_name = schema.get('type', 'object')
    if not schematizable(obj, type_name) and default:
        d = default(obj, type_name)
        if d is not None:
            obj = d
    if 'properties' in schema:
        result = {}
        is_dict = isinstance(obj, dict)
        required_keys = schema.get('required', [])
        for key, sub_schema in schema.get('properties', {}).items():
            _keys = keys + [key]
            required = key in required_keys
            exist = is_dict and key in obj or hasattr(obj, key)
            if not required and not exist:
                continue
            if required and not exist:
                raise SchematizeError(f'key({_keys}) is required')
            try:
                if is_dict:
                    val = obj.get(key)
                else:
                    val = getattr(obj, key)
            except AttributeError:
                raise SchematizeError(f'can\'t get "{obj}" value on key({_keys})')
            result[key] = schematize(sub_schema, val, default, _keys)
        return result
    elif 'items' in schema:
        try:
            return [schematize(schema.get('items', {}), item, default, keys + ['[]']) for item in obj]
        except TypeError:
            raise SchematizeError(f'not iteralbe object "{type(obj)}" on key({keys})')
    else:
        val = obj if obj is not None else schema.get('default', None)
        if val is not None and not schematizable(val, type_name):
            raise SchematizeError(f'not schematizable type: {type(val)} to {type_name} on key({keys})')
        return val



import hmac
import hashlib
import time
from base64 import urlsafe_b64decode
from base64 import urlsafe_b64encode

import orjson


def hs256(key: bytes, content: bytes):
    sign = hmac.new(key, content, hashlib.sha256).digest()
    return b64encode(sign)


SUPPORT_ALGORITHMS = {
    'HS256': hs256
}
DEFAULT_ALGORITHMS = 'HS256'
TOKEN_TTL = 8 * 60 * 60


class ExpiredToken(ValueError):
    pass


class InvalidToken(ValueError):
    pass


def b64encode(s: bytes):
    return urlsafe_b64encode(s).rstrip(b'=')


def b64decode(s: str):
    return urlsafe_b64decode(s + '=' * (len(s) % 4))


def loads(key: str, token: str, check_exp=True):
    terms = token.strip().split('.')
    alg = DEFAULT_ALGORITHMS
    if len(terms) < 3:
        raise InvalidToken(f'invalid token `{token}`')
    header, payload, signature = terms[-3:]
    # select alg by header
    if header:
        # raise ValueError
        h = orjson.loads(b64decode(header))
        try:
            alg = h.get('alg', DEFAULT_ALGORITHMS)
        except AttributeError:
            raise InvalidToken(f'bad header `{token}`')

        if alg not in SUPPORT_ALGORITHMS:
            raise InvalidToken(f'unsupported alg `{token}`')

    sign = SUPPORT_ALGORITHMS[alg]

    content = (header + '.' + payload).encode('u8')
    if signature.encode('u8') != sign(key.encode('u8'), content):
        raise InvalidToken(f'invalid signature `{token}`')

    # raise ValueError
    payload = orjson.loads(b64decode(payload))

    # exp check
    if check_exp and int(payload.get('exp', 0)) < time.time():
        raise ExpiredToken(f'expired token `{token}`')

    return payload


def dumps(key: str, payload: dict, dump_header=False):
    header = {'alg': 'HS256', 'typ': 'JWT'}
    header, payload = [b64encode(orjson.dumps(v))
            for v in (header, payload)]
    if not dump_header:
        header = b''
    content = header + b'.' + payload
    sign = hs256(key.encode('u8'), content)
    token = content + b'.' + sign
    return token.decode()


def build(
        key,
        user_id,
        refresh_token,
        scopes=None,
        data=None,
        ttl=TOKEN_TTL,
    ):
    payload = {
        'sub': user_id,
        'exp': int(time.time()) + ttl
    }
    if data:
        payload['data'] = data
    if scopes:
        payload['scopes'] = scopes
    return {
        'access_token': dumps(key, payload),
        'token_type': 'Bearer',
        'expires_in': ttl,
        'refresh_token': refresh_token,
    }


class JWT:

    def __init__(self, key):
        self.key = key

    def loads(self, token, check_exp=True):
        return loads(self.key, token, check_exp)

    def dumps(self, payload, dump_header=False):
        return dumps(self.key, payload, dump_header)

    def build(
        self,
        user_id,
        refresh_token,
        scopes=None,
        data=None,
        ttl=TOKEN_TTL,
    ):
        return build(self.key, user_id, refresh_token, scopes, data, ttl)




from ..future import singledispatchmethod
from .authorizer import Authorizer
from .validator import Validator
from .operator import Operator
from ..trie import PathTrie
from ..exception import NotFound
from ..exception import PathNotFound
from ..exception import MethodNotAllowed


class Router(object):

    def __init__(self):
        self.trie = PathTrie()

    def add(self, expression):
    	self.trie.add(expression)

    def _set(self, k, v, expression, operation):
        self.trie[expression].setdefault(operation, {})[k] = v

    @singledispatchmethod
    def set(self, v, expression, operator):
        raise NotImplementedError()

    @set.register(Authorizer)
    def _(self, authorizer, expression, operation):
        self._set('authorizer', authorizer, expression, operation)

    @set.register(Validator)
    def _(self, validator, expression, operation):
        self._set('validator', validator, expression, operation)

    @set.register(Operator)
    def _(self, operator, expression, operation):
        self._set('operator', operator, expression, operation)

    def route(self, request):
        expression, sections, operations = self.trie.match(request.path)
        request.environ['fermata.expression'] = expression
        request.environ['fermata.sections'] = sections

        if expression is None:
            raise PathNotFound(f'Path Not Found: `{request.path}`')

        try:
            data = operations[request.method]
        except KeyError:
            if request.method == 'head' and 'get' in operations:
                request.environ['fermata.operation'] = 'get'
                data = operations[request.operation]
            elif request.method in ('get', 'head'):
                raise NotFound('Method Not Found')
            else:
                raise MethodNotAllowed()

        authorizer = data.get('authorizer')
        validator = data.get('validator')
        operator = data.get('operator')

        return authorizer, validator, operator

    def __iter__(self):
        yield from self.trie



import typing


def split(expression):
    parts = expression.strip().strip('/').split('/')
    parts = [] if parts == [''] else parts
    return parts


class PathTrie(object):

    def __init__(self, expressions: typing.Sequence[str] = None) -> None:
        self.root = {}
        if expressions:
            for p in expressions:
                self.add(p)

    def __getitem__(self, expression):
        parts = split(expression)
        node = self.root
        for part in parts:
            if part[0] == '{' and part[-1] == '}':
                part = '/*/'
            node = node[part]
        return node['/data/']

    def __iter__(self):
        nodes = [self.root]
        while nodes:
            current = nodes.pop()
            nodes = [v for k, v in current.items()
                     if k == '/*/' or not k.startswith('/')] + nodes
            if '/expression/' in current:
                yield current['/expression/'], current['/data/']

    def add(self, expression: str) -> None:
        parts = split(expression)
        node = self.root
        sections = []
        for part in parts:
            if part[0] == '{' and part[-1] == '}':
                sections.append(part[1:-1])
                node = node.setdefault('/*/', {})
            else:
                node = node.setdefault(part, {})
        node['/expression/'] = expression
        node['/sections/'] = sections
        node['/data/'] = {}

    def match(self, path: str) -> typing.Tuple[str, dict]:
        parts = split(path)
        length = len(parts)
        searches = [(self.root, 0, [])]  # (node, depth, matched_sections)

        while searches:
            node, depth, sections = searches.pop(0)
            # found it! each parts were matched
            if depth == length and '/sections/' in node:
                return node['/expression/'], dict(
                    zip(node['/sections/'], sections)), node['/data/']
            # wont match
            if depth >= length:
                continue
            # search next level
            part = parts[depth]
            # full match search first
            if part in node:
                searches.append((node.get(part), depth + 1, sections))
            # wildcard match also acceptable
            if '/*/' in node:
                searches.append((node.get('/*/'), depth + 1, sections + [part]))

        return None, {}, None



from .. import loader
from ..exception import BadOperationId


def default_operation_id(path_exception, operation):
    terms, by, use_by = [], [], False
    for part in path_exception.split('/'):
        if len(part) > 2 and part[0] == '{' and part[-1] == '}':
            by.append(part[1:-1])
            use_by = True
        elif part:
            terms.append(part)
            by.append('')
    if use_by:
        operation = '_'.join([operation, 'by'] + by)
    operation_id = '.'.join([''] + terms + [operation])
    return operation_id


class Operator:

    def __init__(self, operation_id, default_package=None, schema=None):
        if default_package and operation_id.startswith('.'):
            operation_id = default_package + operation_id
        self.operation_id = operation_id
        self.schema = schema

    def load(self):
        self.operate = loader.load(self.operation_id)
        return self

    def operate(self, request, **kwargs):
        return self.load().operate(request, **kwargs)

    def __call__(self, request, **kwargs):
        return self.operate(request, **kwargs)





class Specification:
    '''An OpenAPI specification dict wrapper, for getting value 
       by key path recurrently, and load JSON $ref if necessary.'''

    def __init__(self, spec, refs=None):
        self.data = spec
        if refs is None:
            refs = {}
        self.refs = {**refs, '': self.data}
        self.deref(self.data)

    def deref(self, node, from_uri=''):
        stack = [(node, from_uri)]
        objects = set()
        while len(stack) > 0:
            node, from_uri = stack.pop(0)
            if isinstance(node, dict):
                keys = node.keys()
            elif isinstance(node, list):
                keys = range(len(node))
            else:
                continue

            for k in keys:
                uri = from_uri
                _values = set()

                while isinstance(node[k], dict) and '$ref' in node[k]:
                    value, uri = self._ref(node[k]['$ref'], from_uri=uri)
                    if id(value) in _values:
                        raise RecursionError()
                    _values.add(id(value))
                    node[k] = value

                if isinstance(node[k], (dict, list)) and id(node[k]) not in objects:
                    objects.add(id(node[k]))
                    stack += [(node[k], uri)]

    @classmethod
    def ref2keys(cls, ref, from_uri=''):
        if ref.find('#') == -1:
            ref = '#' + ref
        uri, fragment = ref.split('#', 1)
        return uri or from_uri, tuple(k.replace('~1', '/').replace('~0', '~')
                for k in fragment.strip('/').split('/'))

    def _getitem(self, keys, uri=''):
        node = self.refs[uri]
        for k in keys:
            if isinstance(node, list) or isinstance(node, dict) and k not in node:
                try:
                    k = int(k)
                except ValueError:
                    pass
            node = node[k]
        return node

    def __getitem__(self, keys):
        return self._getitem(keys)

    def _ref(self, ref, default=None, from_uri=''):
        uri, keys = self.ref2keys(ref, from_uri)
        try:
            v = self._getitem(keys, uri)
        except (KeyError, IndexError, TypeError):
            return default, ''
        v = v if v is not None else default
        return v, uri

    def ref(self, ref, default=None):
        return self._ref(ref, default)[0]

    def items(self, expression):
        _, parts = self.ref2keys(expression)
        results = []
        params = []
        keys = []
        queue = [(self.data, parts, params)]
        while len(queue) > 0:
            node, parts, params = queue.pop(0)
            part, parts = parts[0], parts[1:]
            wildcard = len(part) >= 3 and part[0] == '{' and part[-1] == '}'
            is_list = isinstance(node, list)

            if not wildcard:
                if is_list or isinstance(node, dict) and part not in node:
                    try:
                        part = int(part)
                    except ValueError:
                        pass
                try:
                    node = { part: node[part] }
                except (IndexError, KeyError, TypeError):
                    continue
                is_list = False

            items = enumerate(node) if is_list else node.items()

            for k, v in items:
                p = params + [(part[1:-1], k)] if wildcard else params
                if parts:
                    queue.append((v, parts[:], p))
                else:
                    results.append((dict(p), v))

        return results


def deref(spec, refs=None):
    return Specification(spec, refs).data



import setuptools


with open("README.md") as f:
    # long_description = f.read()
    long_description = 'just for test'

setuptools.setup(
    name="fermata",
    version='0.0.7',
    author="rejown",
    author_email="rejown@gmail.com",
    description="A WSGI Server Implementation of OpenAPI 3.0 in Python3.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://fenda.io/ff/fermata",
    packages=['fermata', 'fermata.openapi', 'fermata.ext.apm'],
    python_requires='>=3.7',
    install_requires=[
        'jsonschema',
        'orjson',
    ],
    tests_require=[
        'pytest',
    ],
)



import importlib
import typing
import types
from functools import reduce


def import_module(operation_id: str) -> types.ModuleType:
    # if '.' not exists, raise ValueError
    end = operation_id.rindex('.')
    while end > 0:
        try:
            module = importlib.import_module(operation_id[:end])
            return module, operation_id[end + 1:]
        except ImportError:
            # if no module loaded, raise ValueError
            end = operation_id.rindex('.', 0, end)

    raise ValueError(f'operation_id error: "{operation_id}"')


def load(operation_id: str) -> typing.Callable:
    # load module
    mod, obj_path = import_module(operation_id)
    # load func from mod.attr
    return reduce(getattr, obj_path.split('.'), mod)


def preload(operation_ids: typing.Sequence[str]) -> None:
    for op in operation_ids:
        import_module(op)
